#!/usr/bin/perl -w
#< header
use strict;
use Carp;
use Data::Dumper;
use Getopt::Long;
use BerkeleyDB;
use MLDBM qw(BerkeleyDB::Hash Storable);
#>
#< documentation
use constant USAGE => <<"EOT"
usage: extractpernid.pl 05.nuggets.list relate.results/RUN-* > nuggets-by-id.txt
EOT
    ;
=pod

=head1 SYNOPSYS

usage: extractpernid.pl 05.nuggets.list relate.results/RUN-* > nuggets-by-id.txt

=head1 DESCRIPTION

Gather all the nuggets for a given nugget id, and then show all
nuggets in the qid, but not assigned any nuggets.

=head1 NOTES

=head1 SEE ALSO

nuggeteer.pl

=head1 BUGS

=cut
    ;
#> 
#< defaults and user interaction
my $tmpdb = "extractpernid-$$.db";

GetOptions ("help" => sub {exec qq(perldoc $0);},
	    "Version" => sub {print VERSION;exit 0;},
	    
	    "db=s" => \$tmpdb,
	    ) or die USAGE;
die USAGE unless @ARGV;

#>

#< nuggetlistfile

my %nuggets = ();
my %vitoks = ();
my $nuggetlistfile = shift @ARGV;
my %questions = ();
open(NUGGETS, "<", $nuggetlistfile)
    or die "cannot read [$nuggetlistfile]: $!\n";
while (<NUGGETS>) {
    next unless /\S/;
    if (/^Qid ([\d.]+):\s+(.*)/) {
	$questions{$1} = $2;
	$vitoks{$1} = { 0 => "none" };
	next;
    }
    chomp;
    my ($qid, $nid, $vitok, $string) 
	= (/^([\d\.]+)\s(\d+)\s+(vital|okay)\s+(.*)/);
    $nuggets{$qid}||={};
    die "duplicate nugget [$qid $nid] at $nuggetlistfile:$.\n"
	if exists $nuggets{$qid}{$nid};
    $nuggets{$qid}{$nid} = $string;
    $vitoks{$qid} ||={};
    die "duplicate vitok [$qid $nid] at $nuggetlistfile:$.\n" 
	if exists $vitoks{$qid}{$nid}; # should really never happen.
    $vitoks{$qid}{$nid} = $vitok;
}
print STDERR "read nuggets for ".scalar(keys %nuggets)." qids.\n";

#>

#< ndb
my $ndb;
if (-e $tmpdb) {
    $ndb = readdb($tmpdb);
} else {
    $ndb = createdb($tmpdb);
}
#>

#< judgementfiles

warn "No judgement files.  Reading from stdin...\n" unless @ARGV;
my %lastq = ();
my %lastd = ();
my %lastg = ();
my $lastqid = "";
foreach my $jfile (@ARGV) {
    open (J, "<", $jfile) 
	or die "cannot read [$jfile]: $!\n";
    $jfile =~ s|.*/||;
    while (<J>) {
	chomp;
	if (/^([\d\.]+)\s\S+\s(\d+)\s(\S+)\s(.*)$/) { # string line
	    my ($qid, $sid, $docid, $string) = ($1, $2, $3, $4, $5);
	    $string =~ s/\s+/ /; $string =~ s/^\s//; $string =~ s/ $//;
	    $lastq{$sid} = lc($string);
	    $lastd{$sid} = $docid;
	} elsif (/^\s*$/) { # empty
	    if ($lastqid) {
		my $qid = $lastqid;
		my $none = $ndb->{"$qid;0"};
		$none ||= {};
		foreach my $sid (keys %lastq) {
		    # remove sentences that had nuggets:
		    next if $lastg{$sid}; 
		    $none->{qq($jfile\t$lastd{$sid}\t$lastq{$sid})} = 1;
		}
		$ndb->{"$qid;0"} = $none;
		%lastq = (); %lastd = (); %lastg = (); $lastqid = "";
		print STDERR ".";
	    }
	} elsif (/^Nuggets assigned to items for question ([\d\.]+):$/) { 
	    $lastqid = $1; # trigger falses even if there were no good nuggets
	} elsif (/^Submission /) { # worthless;
	} elsif (/^([\d\.]+)\s\S+\s(\d+)\s(\d+)$/) { # nugget assignment
	    my ($qid, $sid, $nid) = ($1, $2, $3);
	    die "unknown qid nid pair [$qid $nid] at $.\n"
		unless exists $nuggets{$qid} and exists $nuggets{$qid}{$nid};
	    die "no string for $qid.sid=$sid in ".Dumper(\%lastq, \%lastd)
		unless exists $lastq{$sid} and exists $lastd{$sid};
	    my $string = $lastq{$sid};
	    my $docid = $lastd{$sid};
	    $lastg{$sid}++;
	    my $nugg = $ndb->{"$qid;$nid"};
	    $nugg ||= {};
	    $nugg->{qq($jfile\t$docid\t$string)}=1;
	    $ndb->{"$qid;$nid"} = $nugg;
	} else {
	    die "unrecognized judgement line $.: [$_]\n";
	}
    }
}
#>

#< dump the database
foreach my $rid (sort { my ($aq, $an) = split /;/, $a; 
			my ($bq, $bn) = split /;/, $b;
			($aq <=> $bq) or ($an <=> $bn); }
		 keys %$ndb) {
    my ($qid, $nid) = split /;/, $rid;
    my $responses = $ndb->{$rid};
    foreach my $response (keys %$responses) {
	#my ($jfile, $docid, $string) = split /\t/, $response;
	print qq($qid\t$nid\t$vitoks{$qid}{$nid}\t$response\n);
    }
}
#>


#< readdb
sub readdb {
    my $name = shift;
    my %h = ();
    tie (%h, "MLDBM", {-Filename=> "$name",
                       -Flags => DB_RDONLY,
                       -Cachesize => 100000000}) 
        or die "Cannot open file $name: $! $BerkeleyDB::Error\n";
    return \%h;
}
#> 
#< createdb
sub createdb {
    my $file = shift;
    my %h = ();
    unlink($file);
    tie(%h, "MLDBM",
	-Filename => $file,
	-Flags    => DB_CREATE)
	or die "Cannot open file $file: $! $BerkeleyDB::Error\n" ;
    return \%h;
}
#> 

1;
