#!/usr/bin/perl -w
#< header
use strict;
use Carp;
use Data::Dumper;
use Getopt::Long;
use BerkeleyDB;
use MLDBM qw(BerkeleyDB::Hash Storable);
use porter;
use Idf qw(fmeas); # and idfs and words and nonstopwords, but these are used by reference and should not be imported. 
#>
#< documentation
use constant VERSION => q$Id: nuggeteer.pl,v 1.5 2006/01/03 21:48:01 gremio Exp $.$/;
our $VERSION = sprintf "%d.%03d", VERSION =~ /(\d+)/g;
use constant USAGE => <<"EOT"
usage: nuggeteer.pl 05.nuggets.list relate.results/RUN-* submittable
EOT
    ;
=pod

=head1 NAME

Nuggeteer - automatic estimated nugget-based evaluation

=head1 SYNOPSYS

nuggeteer.pl 05.nuggets.list relate.results/RUN-* submittable

=head1 DESCRIPTION

Use the RUN-X judgement files for definition or relationship questions
to rank a new run (or to rank one of those judgement files for
testing).

Output format by default is tab-delimited, and column headers are
printed in the first line.  This is intended to make it easy to do
further numerical analysis on a per-question basis.


--help      This message
--Version   Print the CVS version string and exit

C<Output>: 

--outeval   give output in trec.eval format instead of tab-delimited.
--prec=s    a printf format string for printing real output values.  %5.3f
--quiet     normally dots and other indicators go to stderr to indicate
            a judgement qid read (.), or a response recognized as good (+),
            as not good (-), guessed as good (*) or guess not good (x).
            With --quiet, don't show these. 

--unknownsfile u
            Print an assessment-formatted file u containing all of the 
            responses (and their probably assignments) that were not 
            seen in the training data -- those where it would guess.

--jdb dbfile
            Save the n-gram classifiers after learning them from the 
            judgement files, or use them if the file already exists. 


C<Parameters>:

--beta=f    beta to use in f-measure between precision and recall.

--max-rank n
            Score as if only the top n responses were in the response file.

--threshold t
            t may be a number between zero and one inclusive, and recalls 
            above this value will be accepted.

--descriptions-only  vs.
--judgements-too
            Learn n-gram classifiers only from the nugget descriptions, 
            or also from the judgements assigned to a particular nugget.

--ngrams n
            Build n-gram classifiers using n-grams of length 1 .. n

--idf-weighting  vs.
--count-weighting
            The score that each classifier uses is a weighted n-gram recall
            of the n-grams learned during training.  These parameters choose
            the term weighting strategy.  Count weighting assigns equal weight
            to all terms, whereas idf weighting assigns weight proportional
            to the inverse document frequency of each term.  For a bigram or 
            trigram, the weights are summed. 

--remove-stopwords  vs.
--keep-stopwords 
            When learning n-grams and evaluating responses, should we keep
            or remove stopwords.

--stem  vs. 
--no-stem
            When learning n-grams and evaluating responses, should we first
            use the Porter stemmer on each word? 

           
C<Self-Test>:

--exclude!  excludes the testfile from among those to read, if present. 
            Use this to make leave-one-out grading easier.

--ensure-known
            If you don't --exclude, but you're testing a file in your 
            training set, then all responses should be known.  This 
            just makes sure that's the case, dying otherwise, and 
            suppresses default warnings about doing this. 


=head1 NOTES

=head1 SEE ALSO

=head1 BUGS

=cut
    ;
#> 
#< defaults and user interaction
my $tmpdb = "/tmp/autonuggeteer-$$.db";

my $outputformat = "tab";
my $onfmt = "%5.3f";
my $beta=3;
my $excl=0;
my $maxrank=0;
my $unknownsfile;
my $jdb = "";
my $ngrams = 0;
my $threshold = 0.25;
my $descrip_only = 1;
my $stem = 0;
my $quiet = 0;
my $term_weighting = "idf";
my $stopwords = "keep";
GetOptions ("help" => sub {exec qq(perldoc $0);},
	    "Version" => sub {print VERSION;exit 0;},
	    "quiet" => \$quiet,
	    "outeval" => sub { $outputformat = "trec.eval"; },
	    "prec=s" => \$onfmt,
	    "unknownsfile=s" => \$unknownsfile,
	    "jdb=s" => \$jdb,

	    "exclude!" => \$excl,
	    "maxrank=i" => \$maxrank,

	    "threshold=f" => \$threshold,
	    "beta=f" => \$beta,

	    "descriptions-only!" => \$descrip_only,
	    "judgements-too!" => sub {$descrip_only = not shift;},
	    "learn-judgements!" => sub {$descrip_only = not shift;},

	    "ngrams=i" => \$ngrams,
	    "nograms" => sub{ $ngrams = 0; },
	    "unigrams" => sub { $ngrams = 1; },
	    "bigrams" => sub { $ngrams = 2; }, 
	    "trigrams" => sub { $ngrams = 3; },

	    "idf-weighting" => sub { $term_weighting = "idf"; },
	    "count-weighting" => sub { $term_weighting = "count"; },

	    "remove-stopwords" => sub { $stopwords = "remove"; },
	    "keep-stopwords" =>  sub { $stopwords = "keep"; },

	    "stem!" => \$stem,
	    ) or die USAGE;
unless (@ARGV > 1) {
    die USAGE;
}

$descrip_only = 1 unless $ngrams;
my $learn_judgements = not $descrip_only;
$jdb ||= $tmpdb;

my $testfile = pop @ARGV;
open(TESTFILE, "<", $testfile) 
    or die "cannot read [$testfile]: $!\n";

my $term_weight_sub = undef;
if ($term_weighting eq "idf") {
    $term_weight_sub = \&Idf::idfs;
    die "Idf module doesn't seem to have a database.\n"
	if &{$term_weight_sub}("the") eq &{$term_weight_sub}("finish");
} elsif ($term_weighting eq "count") {
    $term_weight_sub = sub { return scalar(@_); };
    die "Count weighting is doing something strange.\n"
	unless (&{$term_weight_sub}("the") == 1 and 
		&{$term_weight_sub}(qw(the fox)) == 2 and 
		&{$term_weight_sub}(qw(the quick fox)) == 3 and
		&{$term_weight_sub}() == 0);
} else {
    die "Unrecognized term weighting scheme: [$term_weighting] (want idf or count)\n";
}

my $word_selection_sub = undef;
if ($stopwords eq "keep") {
    $word_selection_sub = \&Idf::words;
} elsif ($stopwords eq "remove") {
    $word_selection_sub = \&Idf::nonstopwords;
} else {
    die "Unrecognized stopword policy: [$stopwords] (want keep or remove)\n";
}

#>

#< nuggetlistfile

my %nuggets = ();
my %vitoks = ();
my $nuggetlistfile = shift @ARGV;
my %questions = ();
open(NUGGETS, "<", $nuggetlistfile)
    or die "cannot read [$nuggetlistfile]: $!\n";
while (<NUGGETS>) {
    next unless /\S/;
    if (/^Qid ([\d.]+):(\s+'other' question for target)? (.*)/) {
	$questions{$1} = $3;
	next;
    }
    chomp;
    my ($qid, $nid, $vitok, $string) 
	= (/^([\d\.]+)\s(\d+)\s+(vital|okay)\s+(.*)/);
    $nuggets{$qid}||={};
    $vitoks{$qid} ||={};
    if (exists $nuggets{$qid}{$nid}) {
	$nuggets{$qid}{$nid}.="\t".$string;
	die "mismatched vitok [$qid $nid] at $nuggetlistfile:$.\n" 
	    if exists $vitoks{$qid}{$nid} and $vitoks{$qid}{$nid} ne $vitok; 
    } else {
	$nuggets{$qid}{$nid}=$string;
	$vitoks{$qid}{$nid} = $vitok;
    }
}
print STDERR "read nuggets for ".scalar(keys %nuggets)." qids.\n" unless $quiet;

#>
#< --exclude some judgement files
my @jfiles = @ARGV;
if (1<@ARGV and $excl) {
    my $testre = $testfile;
    $testre =~ s/\.[^\.\/]*$//;
    @jfiles = grep !/^$testre/, @jfiles;
    @ARGV = @jfiles;
}
#>

#< ndb
my $ndb;
if (-e $jdb) {
    $ndb = readdb($jdb);
    if (defined $ndb->{_settings}) {
	if ($ndb->{_settings}{learn_judgements} eq $learn_judgements and
	    $ndb->{_settings}{stem} eq $stem and
	    $ndb->{_settings}{ngrams} == $ngrams and 
	    $ndb->{_settings}{stopwords} eq $stopwords and 
	    $ndb->{_settings}{term_weighting} eq $term_weighting and 
	    $ndb->{_settings}{nuggetlistfile} eq $nuggetlistfile 
	    ) {
	    my @dbjfiles = sort @{$ndb->{_settings}{jfiles}};
	    my $okay=1;
	    foreach my $jfile (sort @jfiles) {
		my $dbj = shift @dbjfiles;
		unless ($dbj eq $jfile) { $okay=0; last; }
	    }
	    die "[$jdb] was built from a different set of judgement files.\n"
		unless $okay;
	} else {
	    die "[$jdb] was built with different settings: ".
		Dumper($ndb->{_settings});
	}
    } else {
	die "[$jdb] is incomplete.  Perhaps its build was interrupted?\n";
    }
    $learn_judgements = 0;
} else {
    $ndb = createdb($jdb);
    if ($ngrams) {
	foreach my $qid (keys %nuggets) {
	    learn($ndb, $qid, -1, $questions{$qid});
	    foreach my $nid (keys %{$nuggets{$qid}}) {
		foreach my $ntxt (split /\t/, $nuggets{$qid}{$nid}) {
		    learn($ndb, $qid, $nid, $ntxt);
		}
	    }
	}
    }
}
#>

#< print settings
print qq(\# this = [$0]\n);
print qq(\# starting at: ).localtime()."\n";
print qq(\# host = $ENV{HOST}\n) if defined $ENV{HOST};
print qq(\# user = $ENV{USER}\n) if defined $ENV{USER};
print qq(\# nuggetlistfile = [$nuggetlistfile]\n);
print qq(\# testfile = [$testfile]\n);
print qq(\# judgement files = [). join(", ",@jfiles).qq(]\n);
print qq(\# beta = $beta\n);
print qq(\# maxrank = $maxrank\n);
print qq(\# learn from judgement files = $learn_judgements\n);
print qq(\# ngrams = $ngrams\n);
print qq(\# stopwords policy = $stopwords\n);
print qq(\# term weighting = $term_weighting\n);
print qq(\# global threshold = $threshold\n);
print qq(\# stemming = $stem\n);
print qq(\# database = [$tmpdb]\n);
print qq(\# output unknowns to = [$unknownsfile]\n) if defined $unknownsfile;
print qq(\# result output format = [$outputformat]\n);
#>
#< judgementfiles

warn "No judgement files.  Reading from stdin...\n" unless @ARGV;
my %truth = ();
my %false = ();
my %lastq = ();
my %lasts = ();
my %lastg = ();
my $lastqid = "";
my $ntruth = 0;
my $nfalse = 0;
while (<>) {
    chomp;
    if (/^([\d\.]+)\s\S+\s(\d+)\s(\S+)\s(.*)$/) { # string line
	my ($qid, $sid, $docid, $string) = ($1, $2, $3, $4, $5);
	$string =~ s/\s+/ /; $string =~ s/^\s//; $string =~ s/ $//;
	$lastq{$sid} = lc($string);
	$lasts{$sid} = $docid;
    } elsif (/^\s*$/) { # empty
	if ($lastqid) {
	    my $qid = $lastqid;
	    $false{$qid} ||= {};
	    foreach my $sid (keys %lastq) {
		next if $lastg{$sid}; # skip sentences that had nuggets
		$false{$qid}{$lastq{$sid}} = $lasts{$sid};
		learn($ndb, $qid, 0, $lastq{$sid}) if $learn_judgements;
		$nfalse++;
	    }
	    %lastq = (); %lasts = (); %lastg = (); $lastqid = "";
	    print STDERR "." unless $quiet;
	}
    } elsif (/^Nuggets assigned to items for question ([\d\.]+):$/) { 
	$lastqid = $1; # trigger falses even if there were no good nuggets
    } elsif (/^Submission /) { # worthless;
    } elsif (/^([\d\.]+)\s\S+\s(\d+)\s(\d+)$/) { # nugget assignment
	my ($qid, $sid, $nid) = ($1, $2, $3);
	die "unknown qid nid pair [$qid $nid] at $.\n"
	    unless exists $nuggets{$qid} and exists $nuggets{$qid}{$nid};
	die "no string for $qid.sid=$sid in ".Dumper(\%lastq, \%lasts)
	    unless exists $lastq{$sid} and exists $lasts{$sid};
	my $string = $lastq{$sid};
	my $docid = $lasts{$sid};
	$lastg{$sid}++;
	$truth{$qid} ||= {};
	if (exists $truth{$qid}{$string}) {
	    push @{$truth{$qid}{$string}}, $nid, $docid;
	} else {
	    $truth{$qid}{$string} = [ $nid, $docid ];
	}
	learn($ndb, $qid, $nid, $string) if $learn_judgements;
	$ntruth++;
    } else {
	warn "unrecognized judgement line $.: [$_]\n";
    }
}
print "\# read $ntruth positive and $nfalse negative examples\n";

#>

#< count the testfile
my %results = ();
my %lengths = ();
my %support = ();
my %unknown = ();
my $lruntag = "";
my $ncorrect = 0; 
my $icorrect = 0; 
my $nwrong = 0;
my $nunknown = 0;
my %ngrams = ("" => "");
my %seen = ();
my $unknass = "";
if (defined $unknownsfile) {
    open(UNKOUT, ">", $unknownsfile) or do {
	warn "cannot write [$unknownsfile]: $!\n";
	undef $unknownsfile;
    }
}
$lastqid = 0;
my %ourass = ();
while (<TESTFILE>) {
    my ($qid, $runtag, $sid, $docid, $string);
    #< if this is actually a training file, keep tabs on how well we're doing
    if (/^Nuggets assigned to items for question ([\d\.]+):$/) {
	$qid = $1;
	$_ = <TESTFILE>;
	my $tru = 0;
	my $obs = 0;
	foreach my $k (keys %ourass) { $obs += keys(%{$ourass{$k}}); }
	my $cor = 0;
	while (!/^$/) {
	    chomp;
	    my ($qid, $runtag, $osid, $nid) = split /\s+/;
	    $tru++;
	    $cor++ if exists $ourass{$osid} and exists $ourass{$osid}{$nid};
	    $_ = <TESTFILE>;
	}
	%ourass = ();
	$sid=0;
        my $recall = ($tru)?$cor/$tru:1;
        my $precision = ($obs)?$cor/$obs:$recall;
        print join("\t", "\#EVAL", $qid, $tru, $obs, $cor, 
                   $precision, $recall, fmeas($precision, $recall, 1)).$/;
    }
    #>

    #< read the above, or go on

    chomp;
    if (/^([\d\.]+)\t(\S+)\t(\S+)\t(.*)$/) {
	($qid, $runtag, $docid, $string) = split /\t/;
	$sid++;
    } elsif (/^([\d\.]+)\s+(\S+)\s+(\d+)\s+(\S+)\s+(.*)$/) { # string line
	($qid, $runtag, $sid, $docid, $string) = ($1, $2, $3, $4, $5);
    } else {
	# skip it.
	next;
    }
    next if $docid =~ /^\d+$/ and $string =~ /^\d+$/; #don't read nugget assignments
    next if $maxrank and (++$seen{$qid}) > $maxrank;

    #>
    $lruntag = $runtag;
    $string = join(" ", split /\s+/, $string); # just to make sure.

    my $wordlen = $string; $wordlen =~ s/\s*//g; $wordlen = length $wordlen;
    $lengths{$qid}||=0; $lengths{$qid} += $wordlen;

    my $origstring = $string;
    $string =~ s/\s+/ /; $string =~ s/^\s//; $string =~ s/ $//; 
    $string = lc($string);

    if (exists $truth{$qid}{$string}) {
	my %nids = @{$truth{$qid}{$string}};
	foreach my $nid (keys %nids) {
	    #< record the matched nuggets
	    my $tdocid = $nids{$nid};
	    next if exists $support{$qid} and exists $support{$qid}{$nid};

	    $results{$qid}||={};
	    $results{$qid}{$nid} = $string;

	    $ourass{$sid}||={};
	    $ourass{$sid}{$nid}++;

	    if ($docid eq $tdocid) {
		# this is still the wrong condition because i only
		# allow one tdocid, which is sort of presumably the
		# latest per string.  I should keep all of them and
		# mark any as supported.

		$support{$qid}||={};
		$support{$qid}{$nid} = $docid;
		$ncorrect++;
	    }
	    #>
	}
	print STDERR "+" unless $quiet;
	$icorrect++;
    } elsif (exists $false{$qid}{$string}) {
	print STDERR "-" unless $quiet;
	$nwrong++;
	# do nothing.
    } else {
	$unknown{$qid}||=0;
	$unknown{$qid}++;
	#< print this item to the unknowns file

	#warn "cannot find $qid [$string]\n";
	if ($unknownsfile) {
	    if ($unknown{$qid} == 1) {
		if ($lastqid) {
		    print UNKOUT "\nNuggets assigned to items for question $lastqid:\n";
		    print UNKOUT $unknass;
		    print UNKOUT qq(\n\n\n);
		    $unknass = "";
		}		
		print UNKOUT "Submission for question $qid with items numbered:\n";
	    }
	    print UNKOUT qq($qid $lruntag-UNK $unknown{$qid} $docid $origstring\n);
	    $lastqid = $qid;
	}

	#>
	my $judgement = judge($ndb, $qid, $string);
	#< record judgement
	if ($judgement and @$judgement) {
	    foreach my $nid (@$judgement) {
		next if exists $support{$qid} and exists $support{$qid}{$nid};
		
		$ourass{$sid}||={};
		$ourass{$sid}{$nid}++;
		$results{$qid}||={};
		$results{$qid}{$nid} = $docid;
		$support{$qid}||={};
		$support{$qid}{$nid} = $docid;
		$unknass .= "$qid\t$lruntag-UNK\t$unknown{$qid}\t$nid\n";

		$ncorrect++
	    }
	    $icorrect++;
	    print STDERR "*" unless $quiet;
	} else {
	    $nunknown++;
	    print STDERR "x" unless $quiet;
	}
	#>
    }    
}
if ($unknownsfile and $unknass) {
    print UNKOUT "\nNuggets assigned to items for question $lastqid:\n";
    print UNKOUT $unknass;
    print UNKOUT qq(\n\n\n);
}
close UNKOUT if $unknownsfile;
print "#found $ncorrect correct, $nwrong wrong, and $nunknown unknown responses.\n";
#>
print STDERR "\n" unless $quiet;

#< save db settings
unless (exists $ndb->{_settings}) {
    if ($jdb eq $tmpdb) {
	undef $ndb;
	unlink $tmpdb;
    } else {
	my %settings = ();
	$settings{learn_judgements} = not $descrip_only;
	$settings{stem} = $stem;
	$settings{ngrams} = $ngrams;
	$settings{stopwords} = $stopwords;
	$settings{term_weighting} = $term_weighting;
	$settings{nuggetlistfile} = $nuggetlistfile;
	$settings{jfiles} = \@jfiles;
	$ndb->{_settings} = \%settings;
    }
}
#>

#< print results
if ($outputformat eq "tab") {
    print "#".join("\t", qw(qid runtag----- vital okay length recall prec f)).$/;
}
foreach my $qid (sort {$a <=> $b} keys %nuggets) {
    my $vital_poss = scalar(grep /vital/, values %{$vitoks{$qid}});
    my $okay_poss =  scalar(grep /okay/, values %{$vitoks{$qid}});
    my $total_poss = $vital_poss+$okay_poss;

    my $vital_match = 0;
    my $okay_match = 0;
    foreach my $nid (keys %{$results{$qid}}) {
	($vitoks{$qid}{$nid} eq "vital")? $vital_match++ : $okay_match++;
    }
    my $total_match = $vital_match+$okay_match;

    my $recall = $vital_match / $vital_poss;

    my $allowance = 100 * $total_match;
    my $length = $lengths{$qid};
    $length ||= 0;
    my $precision = ($length < $allowance) ? 1 
	: (($length) ? (1 - ($length - $allowance) / $length) : 0 );

    my $fmeasure = fmeas($precision, $recall, $beta);

    if ($outputformat eq "trec.eval") {
	print qq(Qid: $qid\t$lruntag\n);
	print qq(\tVital matches:   $vital_match/$vital_poss\n);
	print qq(\tTotal matches:   $total_match/$total_poss\n);
	print qq(\tLength of response: $length\n);
	printf qq(\tRecall: $onfmt\n), $recall;
	printf qq(\tPrecision $onfmt\n), $precision;
	printf qq(\tF (with beta=$beta): $onfmt\n), $fmeasure;
    } else {
	print join("\t", $qid, $lruntag, $vital_match, $okay_match, $length, 
		   sprintf($onfmt, $recall), sprintf($onfmt, $precision), 
		   sprintf($onfmt, $fmeasure)) . $/;
    }
}
#>

# The nugget database looks like this: 
# $ndb = { question-id => { nugget-id => [ nugget-instances ] } }, where
# nugget-instance = { _sum => total-weight, _orig => string, n-gram => weight }
# 
# So we're set to compute recall for each nugget instance by finding 
# matching n-gram instances, and dividing the sum of their weights by
# the sum of the weights in that nugget instance. 
# 
# We can then compute a nugget score from nugget instance scores by 
# multiplying doubts (rather than confidences), so that nugget weight is
#    1 - \sum_{i \in nugget-instances}{ 1 - recall_i }
#
# A special nugget-id "_inform" gives us: 
# $ndb = { question-id => { _inform => { n-gram => { i => 0..1, 
#                                            nugget-id => 1 } } } }

#< \@nids = judge($ndb, $qid, $string) 
sub judge {
    my ($ndb, $qid, $string) = @_;
    return undef unless $ngrams;
    return undef unless "HASH" eq ref $ndb and exists $ndb->{$qid};
    confess "no string?\n" unless $string;
    confess "strange ndp for qid=$qid:".Dumper($ndb->{$qid}) 
	unless "HASH" eq ref $ndb->{$qid};
    
    my @words = grep { s/\W//g; $_; } &{$word_selection_sub}($string);
    @words = map { porter($_) } @words if $stem;    

    my $nuggets = $ndb->{$qid};
    my $inform = $nuggets->{_inform};

    my %scores = ();
    my %explanations = ();
    foreach my $indicator (indicators(@words)) {
	if (exists $inform->{$indicator}) {
	    foreach my $nid (keys %{$inform->{$indicator}}) {
		next if $nid eq "i";
		$scores{$nid} ||= {};
		$explanations{$nid} ||= {};
		my $i=0;
		foreach my $ninstance (@{$nuggets->{$nid}}) {
		    if (exists $ninstance->{$indicator}) {
			$scores{$nid}{$i} += ($ninstance->{$indicator} * 
					      $inform->{$indicator}{i});
			$explanations{$nid}{$i} .= " + $indicator: $ninstance->{$indicator}*$inform->{$indicator}{i}";
		    } else {
			# do nothing, the current instance wasn't here. 
		    }
		    $i++;
		}
	    }
	} else {
	    # can't tell anything from here. 
	}
    }

    # now scores = { nid => { instance => sum-of(matched-indicator-weights 
    #                                            * their-informativenesses) } }
    # so I can give each instance, and thus each nugget, a score. 

    foreach my $nid (keys %scores) {
	# PRODUCT OF DOUBTS: 
#	my $score = 1;
#	foreach my $instance (keys %{$scores{$nid}}) {
#	    $explanations{$nid}{$instance} .= " = $scores{$nid}{$instance}";
#	    $score *= 1 - ($scores{$nid}{$instance} / 
#			   $nuggets->{$nid}[$instance]{_sum});
#	    $explanations{$nid}{$instance} .= "; / $nuggets->{$nid}[$instance]{_sum}";
#	}
#	$scores{$nid} = 1 - $score;

	# MAX: 
	my $score = 0;
	foreach my $instance (keys %{$scores{$nid}}) {
	    my $iscore = ($scores{$nid}{$instance} / 
			  $nuggets->{$nid}[$instance]{_sum});
	    if ($iscore > $score) {
		$score = $iscore;
	       $explanations{$nid}{$instance} .= " = $scores{$nid}{$instance}"
		   . "; / $nuggets->{$nid}[$instance]{_sum} = $iscore";
	    }
	}
	$scores{$nid} = $score;
    }


    my @assigned_nids = ();
    my $rank=0;
    $nuggets{$qid}{0} = "##NONE##";
    my $local_threshold = $threshold;
    $local_threshold = $scores{0} if 
	$scores{0} and $scores{0} > $local_threshold;
    foreach my $nid (sort {$scores{$b} <=> $scores{$a}} keys %scores) { 
	$rank++;
	last unless $scores{$nid} > $local_threshold # cutoff at threshold
	    or ($ENV{DEBUG} and $ENV{DEBUG} > 1);
	print qq(\#DEBUG $qid\t$rank\t$nid\t$scores{$nid}\t$string).
	    qq(\tNUGGET $nid\t$nuggets->{$nid}[0]{_orig}\n);
	last unless $nid # background matched better than later nuggets
	    or $ENV{DEBUG}; 
	push @assigned_nids, $nid if $scores{$nid} > $local_threshold;
	if ($ENV{DEBUG} and $ENV{DEBUG} > 2) {
	    print qq(\#DEBUG score    $nid = $scores{$nid}\n);
	    foreach my $instance ( keys %{$explanations{$nid}} ) {
		$explanations{$nid}{$instance} =~ s/(\d+\.\d\d\d\d)\d+/$1/g;
		print qq(\#DEBUG x        ).$explanations{$nid}{$instance}.$/;
		print qq(\#DEBUG n        $nuggets->{$nid}[$instance]{_orig}\n)
	    }
	    print qq(\n\n);
	}
    }

    return \@assigned_nids;
}
#>
#< learn($ndb, $qid, $nid, $string)
sub learn {
    my ($ndb, $qid, $nid, $string) = @_;
    confess "no ndb!!?\n" unless $ndb and "HASH" eq ref $ndb;
    my $question = $ndb->{$qid}; 
    $question ||= {_inform => {}};
    my $nugget;
    if (exists $question->{$nid}) {
	$nugget = $question->{$nid};
    } else { 
	$nugget = [];
	$question->{$nid} = $nugget;
    }
    my $ninstance = {_sum => 0, _orig => $string};
    my $inform = $question->{_inform};
    my @words = grep { s/\W//g; $_; } &{$word_selection_sub}($string);
    my @indicators = indicators(@words);

    #< update counts and informativeness
    foreach my $indicator (@indicators) {
	my $weight = &{$term_weight_sub}(split /\s/, $indicator);
	$indicator = lc($indicator); # we fold case AFTER we get the idf
	$indicator = stem $indicator if $stem; # we stem AFTER we get the idf
	next if exists $inform->{$indicator}{-1}; # question word

	$ninstance->{$indicator} = $weight;
	$ninstance->{_sum} += $weight;

	$inform->{$indicator} ||= {i => 0};
	if (exists $inform->{$indicator}{$nid}) {
	    # do nothing: no need to update it or update the informativeness
	} else {
	    $inform->{$indicator}{$nid}=1;
	    $inform->{$indicator}{i} = 1 - 
		# number of nuggets this indicator appears in: 
		(scalar(keys(%{$inform->{$indicator}})) - 1) / # -1 for i
		# divided by total number of nuggets for this question:
		(scalar(keys(%$question)) - 1);    # -1 for _inform

	    $inform->{$indicator}{i} = 0
		if (exists $inform->{$indicator}{-1}); # question word
	}
    }
    #>

    push @$nugget, $ninstance;
    $ndb->{$qid} = $question;
}
#>    


#< indicators(@words) -- gather n-and-fewer-gram indicators from @words
sub indicators {
    my @words = @_;
    my @indicators = ();
    foreach my $ng (0 .. $ngrams) {
	foreach my $i ($ng..$#words) {
	    push @indicators, join(" ",@words[($i-$ng) .. $i]);
	}
    }    
    return @indicators;
}
#>

#< readdb
sub readdb {
    my $name = shift;
    my %h = ();
    tie (%h, "MLDBM", {-Filename=> "$name",
                       -Flags => DB_RDONLY,
                       -Cachesize => 100000000}) 
        or die "Cannot open file $name: $! $BerkeleyDB::Error\n";
    return \%h;
}
#> 
#< createdb
sub createdb {
    my $file = shift;
    my %h = ();
    unlink($file);
    tie(%h, "MLDBM",
	-Filename => $file,
	-Flags    => DB_CREATE)
	or die "Cannot open file $file: $! $BerkeleyDB::Error\n" ;
    return \%h;
}
#> 

1;
